import base64
import hmac
import json

from robokassa.types import Hash


class JWT:
    def __init__(
        self, header: dict, payload: dict, signature_key: str, hash: Hash
    ) -> None:
        self._header = header
        self._payload = payload
        self._signature_key = signature_key
        self._hash = hash

    def _encode_to_base64(self, data: bytes) -> str:
        return base64.urlsafe_b64encode(data).decode().rstrip("=")

    def _dict_to_json_string(self, obj: dict) -> str:
        return json.dumps(obj, separators=(",", ":"))

    def _assemble_parts(self, *args) -> str:
        return ".".join(args)

    def _get_hmac(self, message: str) -> hmac.HMAC:
        return hmac.new(
            key=self._signature_key.encode(),
            msg=message.encode(),
            digestmod=self._hash.encrypt(),
        )

    def _encrypt_data(self, message: str) -> str:
        return self._get_hmac(message).digest()

    def create(self) -> str:
        formatted_parts = tuple(
            self._dict_to_json_string(i) for i in (self._header, self._payload)
        )
        encoded_parts = tuple(
            self._encode_to_base64(i.encode()) for i in formatted_parts
        )
        message = self._assemble_parts(*encoded_parts)
        signature = self._encode_to_base64(self._encrypt_data(message))

        return self._assemble_parts(message, signature)

    def __str__(self) -> str:
        return self.create()

    def __repr__(self) -> str:
        return "JWT()"



class RobokassaException(Exception):
    pass


class UnresolvedAlgorithmTypeError(RobokassaException):
    pass


class UnusedStrictUrlParameterError(RobokassaException):
    pass


class UnusedParameterError(RobokassaException):
    pass


class IncorrectUrlMethodError(RobokassaException):
    pass


class RobokassaInterfaceError(RobokassaException):
    pass


class RobokassaParsingError(RobokassaException):
    pass


class RobokassaRequestError(RobokassaException):
    pass



import xml.etree.ElementTree as Et
from json import JSONDecodeError

import httpx

from robokassa.exceptions import RobokassaInterfaceError

correct_keys = {
    "merchant_login": "MerchantLogin",
    "description": "Description",
    "out_sum": "OutSum",
    "signature_value": "SignatureValue",
    "inc_current_label": "IncCurrentLabel",
    "payment_methods": "PaymentMethods",
    "inv_id": "InvId",
    "culture": "Culture",
    "encoding": "Encoding",
    "email": "Email",
    "expiration_data": "ExpirationDate",
    "is_test": "IsTest",
}


def flatten_dict(data, serialize_to_http_naming: bool = False):
    items = {}

    def recurse(current_dict):
        for key, value in current_dict.items():
            if isinstance(value, dict):
                recurse(value)
            else:
                if value is not None:
                    if serialize_to_http_naming:
                        key = correct_keys.get(key, key)
                    items[key] = value

    recurse(data)
    return items


class HttpResponseValidator:
    def __init__(self, response: httpx.Response, in_json: bool = True) -> None:
        self.response = response
        self.in_json = in_json

        if response.status_code != 200:
            raise RobokassaInterfaceError(
                "RobokassaInterface servers are unavailable. Please try again later."
            )

    def xml_to_dict(self, element):
        if len(element) == 0:
            return element.text.strip() if element.text else None
        result = {}
        for child in element:
            tag = child.tag.split("}")[-1]
            child_dict = self.xml_to_dict(child)
            if child.attrib:
                if tag in result:
                    if not isinstance(result[tag], list):
                        result[tag] = [result[tag]]
                    result[tag].append(
                        {**child.attrib, **(child_dict if child_dict else {})}
                    )
                else:
                    result[tag] = {**child.attrib, **(child_dict if child_dict else {})}
            else:
                if tag in result:
                    if not isinstance(result[tag], list):
                        result[tag] = [result[tag]]
                    result[tag].append(child_dict)
                else:
                    result[tag] = child_dict
        return result

    def validate_http_response(self) -> dict:
        if not self.in_json:
            data = self.response.text
            root = Et.fromstring(data)
            return self.xml_to_dict(root)

        try:
            data = self.response.json()
        except JSONDecodeError:
            raise RobokassaInterfaceError("Internal Robokassa server error")
        if data.get("errorCode") == 0:
            return data
        else:
            if data.get("errorCode") is None:
                return data
        raise RobokassaInterfaceError(
            f"Error code: {data.get('errorCode')}, error message: {data.get('errorMessage')}"
        )



from typing import Any, Optional, Union

from robokassa.hash import Hash
from robokassa.types import Signature


class SignatureChecker:
    def __init__(self, hash_: Hash, password1: str, password2: str) -> None:
        self._hash: Hash = hash_
        self._password1 = password1
        self._password2 = password2

    def success_or_fail_url_signature_is_valid(
        self,
        success_signature: str,
        out_sum: Union[str, float, int],
        inv_id: Optional[Union[str, int]] = None,
        **kwargs: Any,
    ) -> bool:
        old_signature = Signature(
            value=success_signature.lower(),
            hash_=self._hash,
        )
        new_signature = Signature(
            out_sum=out_sum,
            inv_id=inv_id,
            additional_params=kwargs,
            password=self._password1,
            hash_=self._hash,
        )

        return old_signature == new_signature

    def result_url_signature_is_valid(
        self,
        result_signature: str,
        out_sum: Union[str, float, int],
        inv_id: Optional[Union[str, int]] = None,
        **kwargs: Any,
    ) -> bool:
        old_signature = Signature(
            value=result_signature.lower(),
            hash_=self._hash,
        )
        new_signature = Signature(
            out_sum=out_sum,
            inv_id=inv_id,
            additional_params=kwargs,
            password=self._password2,
            hash_=self._hash,
        )

        return old_signature == new_signature



import json
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from hmac import compare_digest
from typing import Any, Dict, List, Optional, Union
from urllib.parse import quote

from robokassa.exceptions import UnusedStrictUrlParameterError
from robokassa.hash import Hash


@dataclass
class Signature:
    """ """

    merchant_login: Optional[str] = None

    out_sum: Optional[Union[float, str, int]] = None
    inv_id: Optional[Union[str, int]] = None
    user_ip: Optional[str] = None
    receipt: Optional[dict] = None

    result_url2: Optional[str] = None

    success_url2: Optional[str] = None
    success_url2_method: Optional[str] = None

    fail_url2: Optional[str] = None
    fail_url2_method: Optional[str] = None

    password: Optional[str] = None

    value: Optional[str] = None

    additional_params: Dict[str, Union[int, str, float]] | None = None

    hash_: Optional[Hash] = None

    def __post_init__(self) -> None:
        """
        For create:
        `MerchantLogin:OutSum:InvId:ResultUrl2:SuccessUrl2:SuccessUrl2Method:password1`

        For check:
        `OutSum:InvId:[password 1 or 2]`

        Also, additional params after password if params is containing in Signature.
        """
        if not self._url_data_is_correct(
            self.success_url2, self.success_url2_method
        ) or not self._url_data_is_correct(self.fail_url2, self.fail_url2_method):
            raise UnusedStrictUrlParameterError(
                "If you use success_url2 or fail_url2 don't forget choose"
                "HTTP method for them.\nAvailable HTTP Methods:\n"
                "GET, POST. Use them like a uppercase string"
            )

        if self.value is not None:
            return

        inv_id = "" if self.inv_id is None else self.inv_id
        password = self.password
        hashable_string = self._serialize_string_for_hash(
            self.merchant_login,
            self.out_sum,
            inv_id,
            self.user_ip,
            self.receipt,
            self.result_url2,
            self.success_url2,
            self.success_url2_method,
            self.fail_url2,
            self.fail_url2_method,
            password,
            *self._get_serialized_additional_params(),
        )

        self.value = self._calculate_hash(self.hash_, hashable_string)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Signature):
            raise TypeError("Cannot use this type for signature")

        return compare_digest(other.value, self.value)

    def _get_serialized_additional_params(self) -> list:
        if self.additional_params is None:
            return []
        return sorted(tuple(f"{k}={v}" for k, v in self.additional_params.items()))

    def _serialize_string_for_hash(self, *args) -> str:
        values = tuple(
            str(i) for i in (args) if i is not None and i != "null" and i != -1
        )
        result = ":".join(values)
        return result

    def _calculate_hash(self, hash_: Hash, data: str) -> str:
        return hash_.hash_data(data)

    def _url_data_is_correct(self, url: str, url_method: str) -> bool:
        if (url is not None) or (url_method is not None):
            if (url is None) or (url_method is None):
                return False
        return True


class PaymentMethod(Enum):
    BANK_CARD: str = "BankCard"
    SBP: str = "SBP"


class Culture(Enum):
    RU: str = "ru"
    EN: str = "en"


class HTTPMethod(Enum):
    GET: str = "GET"
    POST: str = "POST"


class InvoiceType(Enum):
    ONE_TIME: str = "OneTime"
    REUSABLE: str = "Reusable"


@dataclass
class RobokassaParams:
    base_url: str = "https://auth.robokassa.ru/Merchant/Index.aspx"
    merchant_login: Optional[str] = None
    out_sum: Optional[Union[float, str, int]] = None
    description: Optional[str] = None
    signature_value: Optional[str] = None
    receipt: Optional[str] = None
    is_test: bool = False

    inc_curr_label: Optional[str] = None
    payment_methods: Optional[List[PaymentMethod]] = None
    merchant_comments: Optional[str] = None
    invoice_type: Optional[InvoiceType] = None
    inv_id: Optional[Union[int, str]] = None
    id: Optional[int] = None
    previous_inv_id: Optional[int] = None
    culture: Optional[Culture] = None
    encoding: Optional[str] = None
    user_ip: Optional[str] = None
    email: Optional[str] = None
    recurring: bool = False
    expiration_date: Optional[datetime] = None
    default_prefix: str = "shp"

    result_url: Optional[str] = None
    success_url: Optional[str] = None
    success_url_method: Optional[HTTPMethod] = None
    fail_url: Optional[str] = None
    fail_url_method: Optional[HTTPMethod] = None

    additional_params: Optional[Dict[str, Any]] = None

    def _encode_receipt(self) -> None:
        encoded = quote(json.dumps(self.receipt, ensure_ascii=False), safe="")
        self.receipt = encoded if encoded != "null" else None

    def __post_init__(self) -> None:
        if not self.additional_params:
            self.additional_params = {}
        self.additional_params = {
            f"{self.default_prefix}_{k}": v for k, v in self.additional_params.items()
        }

        # -------------
        # SERIALIZATION
        # -------------
        self._encode_receipt()
        if self.expiration_date:
            self.expiration_date = self.expiration_date.isoformat()
        if self.payment_methods:
            self.payment_methods = [i.value for i in self.payment_methods]
        if self.culture:
            self.culture = self.culture.value
        if self.success_url_method:
            self.success_url_method = self.success_url_method.value
        if self.fail_url_method:
            self.fail_url_method = self.fail_url_method.value
        if self.invoice_type:
            self.invoice_type = self.invoice_type.value

    def to_dict(self) -> dict:
        """
        Returns dict for POST methods.
        """
        params = (
            ("MerchantLogin", self.merchant_login),
            ("OutSum", self.out_sum),
            ("InvId", self.inv_id),
            ("UserIp", self.user_ip),
            ("Email", self.email),
            ("Description", self.description),
            ("SignatureValue", self.signature_value),
            ("Receipt", self.receipt),
            ("Culture", self.culture),
            ("PaymentMethods", self.payment_methods),
            ("Recurring", self.recurring),
            ("PreviousInvoiceID", self.previous_inv_id),
            ("ResultUrl2", self.result_url),
            ("SuccessUrl2", self.success_url),
            ("SuccessUrl2Method", self.success_url_method),
            ("FailUrl2", self.fail_url),
            ("FailUrl2Method", self.fail_url_method),
            ("ExpirationDate", self.expiration_date),
            ("IsTest", self.is_test),
            *self.additional_params.items(),
        )

        return {
            k: v
            for k, v in params
            if v or (isinstance(v, int) and not isinstance(v, bool) and v == 0)
        }


class PaymentState(Enum):
    INITIATED: int = 5
    CANCELLED: int = 10
    ON_HOLD: int = 20
    PENDING: int = 50
    FAILED: int = 60
    SUSPENDED: int = 80
    COMPLETED: int = 100


@dataclass
class RobokassaResponse:
    url: Optional[str] = None
    params: Optional[RobokassaParams] = None


@dataclass
class PaymentDetails:
    state: Optional[PaymentState] = None
    description: Optional[str] = None
    request_date: Optional[datetime] = None
    state_date: Optional[datetime] = None
    inc_curr_label: Optional[str] = None
    inc_sum: Optional[float] = None
    inc_account: Optional[str] = None
    out_curr_label: Optional[str] = None
    out_sum: Optional[float] = None
    op_key: Optional[str] = None
    bank_card_rrn: Optional[str] = None
    user_fields: Optional[Dict[Any, Any]] = None



from httpx import AsyncClient


class Http(AsyncClient):
    def __init__(self, base_url: str, *args, **kwargs):
        super().__init__(base_url=base_url, *args, **kwargs)


AUTH_BASE_URL = "https://auth.robokassa.ru/Merchant"
SERVICES_BASE_URL = "https://services.robokassa.ru/InvoiceServiceWebApi/api"



import xml.etree.ElementTree as ET
from datetime import datetime
from typing import Optional

from robokassa.connection import Http
from robokassa.exceptions import RobokassaParsingError, RobokassaRequestError
from robokassa.hash import Hash
from robokassa.types import PaymentDetails, PaymentState, Signature


class OperationStateChecker:
    def __init__(self, merchant_login: str, hash: Hash, password_2: str) -> None:
        self._url = "/WebService/Service.asmx/OpStateExt"
        self._merchant_login = merchant_login
        self._hash = hash
        self.__password = password_2

        self._ns = {"ns": "http://merchant.roboxchange.com/WebService/"}

    def _parse_xml(self, text: str) -> ET.Element:
        return ET.fromstring(text.strip())

    def _find_el(self, xml: ET.Element, name: str) -> Optional[ET.Element]:
        return xml.find(name, self._ns)

    def _serialize_xml(self, xml: ET.Element) -> dict:
        return {
            "result": {
                "code": int(self._find_el(xml, "ns:Result/ns:Code").text)
                if self._find_el(xml, "ns:Result/ns:Code") is not None
                else None,
                "description": self._find_el(xml, "ns:Result/ns:Description").text
                if self._find_el(xml, "ns:Result/ns:Description") is not None
                else None,
            },
            "state": {
                "code": int(self._find_el(xml, "ns:State/ns:Code").text)
                if self._find_el(xml, "ns:State/ns:Code") is not None
                else None,
                "request_date": self._find_el(xml, "ns:State/ns:RequestDate").text
                if self._find_el(xml, "ns:State/ns:RequestDate") is not None
                else None,
                "state_date": self._find_el(xml, "ns:State/ns:StateDate").text
                if self._find_el(xml, "ns:State/ns:StateDate") is not None
                else None,
            },
            "info": {
                "inc_curr_label": self._find_el(xml, "ns:Info/ns:IncCurrLabel").text
                if self._find_el(xml, "ns:Info/ns:IncCurrLabel") is not None
                else None,
                "inc_sum": float(self._find_el(xml, "ns:Info/ns:IncSum").text)
                if self._find_el(xml, "ns:Info/ns:IncSum") is not None
                else None,
                "inc_account": self._find_el(xml, "ns:Info/ns:IncAccount").text
                if self._find_el(xml, "ns:Info/ns:IncAccount") is not None
                else None,
                "payment_method": {
                    "code": self._find_el(xml, "ns:Info/ns:PaymentMethod/ns:Code").text
                    if self._find_el(xml, "ns:Info/ns:PaymentMethod/ns:Code")
                    is not None
                    else None,
                    "description": self._find_el(
                        xml, "ns:Info/ns:PaymentMethod/ns:Description"
                    ).text
                    if self._find_el(xml, "ns:Info/ns:PaymentMethod/ns:Description")
                    is not None
                    else None,
                },
                "out_curr_label": self._find_el(xml, "ns:Info/ns:OutCurrLabel").text
                if self._find_el(xml, "ns:Info/ns:OutCurrLabel") is not None
                else None,
                "out_sum": float(self._find_el(xml, "ns:Info/ns:OutSum").text)
                if self._find_el(xml, "ns:Info/ns:OutSum") is not None
                else None,
                "op_key": self._find_el(xml, "ns:Info/ns:OpKey").text
                if self._find_el(xml, "ns:Info/ns:OpKey") is not None
                else None,
                "bank_card_rrn": self._find_el(xml, "ns:Info/ns:BankCardRRN").text
                if self._find_el(xml, "ns:Info/ns:BankCardRRN") is not None
                else None,
            },
            "user_field": [
                {
                    "name": field.find("ns:Name", self._ns).text
                    if field.find("ns:Name", self._ns) is not None
                    else None,
                    "value": field.find("ns:Value", self._ns).text
                    if field.find("ns:Value", self._ns) is not None
                    else None,
                }
                for field in xml.findall("ns:UserField/ns:Field", self._ns)
            ],
        }

    def _handle_result_data(self, code: int) -> None:
        if code == 0:
            return
        elif code == 1:
            raise RobokassaRequestError("Wrong digital signature of request")
        elif code == 2:
            raise RobokassaRequestError("MerchantLogin not found or not activated")
        elif code == 3:
            raise RobokassaRequestError(
                "Information about a spicified InvoiceId not found"
            )
        elif code == 4:
            raise RobokassaRequestError("Found 2 operations with same InvoiceId")
        elif code == 1000:
            raise RobokassaRequestError("Internal Robokassa servers error")
        else:
            raise RobokassaParsingError("Unexpected response code")

    async def get_state(self, http: Http, inv_id: int) -> PaymentDetails:
        request_data = {
            "MerchantLogin": self._merchant_login,
            "InvoiceId": inv_id,
            "Signature": Signature(
                merchant_login=self._merchant_login,
                password=self.__password,
                hash_=self._hash,
                inv_id=inv_id,
            ).value,
        }
        async with http as conn:
            response = await conn.post(self._url, data=request_data)
        try:
            serialized = self._serialize_xml(self._parse_xml(response.text))
        except Exception:
            raise RobokassaParsingError(
                "Cannot parse response from Robokassa servers"
            ) from None

        self._handle_result_data(serialized["result"]["code"])

        state = serialized.get("state")
        info = serialized.get("info")

        return PaymentDetails(
            state=PaymentState(serialized["state"]["code"]),
            description=state.get("description") if state else None,
            request_date=datetime.fromisoformat(state.get("request_date"))
            if state and state.get("request_date")
            else None,
            state_date=datetime.fromisoformat(state.get("state_date"))
            if state and state.get("state_date")
            else None,
            inc_curr_label=info.get("inc_curr_label") if info else None,
            inc_sum=info.get("inc_sum") if info else None,
            inc_account=info.get("inc_account") if info else None,
            out_curr_label=info.get("out_curr_label") if info else None,
            out_sum=info.get("out_sum") if info else None,
            op_key=info.get("op_key") if info else None,
            bank_card_rrn=info.get("ban_card_rrn") if info else None,
            user_fields={k: v for k, v in serialized.get("user_field")}
            if serialized.get("user_field")
            else None,
        )



from datetime import datetime
from typing import List, Optional, Union

from robokassa.connection import AUTH_BASE_URL, SERVICES_BASE_URL, Http
from robokassa.exceptions import (
    IncorrectUrlMethodError,
    RobokassaInterfaceError,
    UnusedParameterError,
    UnusedStrictUrlParameterError,
)
from robokassa.hash import Hash, HashAlgorithm
from robokassa.merchant import OperationStateChecker
from robokassa.payment import LinkGenerator
from robokassa.signature import SignatureChecker
from robokassa.types import (
    Culture,
    HTTPMethod,
    InvoiceType,
    PaymentDetails,
    PaymentMethod,
    RobokassaParams,
    RobokassaResponse,
)


class RobokassaAbstract:
    pass


class BaseRobokassa(RobokassaAbstract):
    def __init__(
        self,
        merchant_login: str,
        password1: str,
        password2: str,
        algorithm: HashAlgorithm = HashAlgorithm.md5,
        is_test: bool = False,
    ) -> None:
        self._merchant_login = merchant_login
        self._password1 = password1
        self._password2 = password2
        self._algorithm = algorithm
        self._is_test = is_test

        self._hash = self._init_hash(self._algorithm)
        self._link_generator = self._init_link_generator()
        self._signature_checker = self._init_signature_checker()
        self._operation_state_checker = self._init_operation_state_checker()

    def _create_http(self, url: str) -> Http:
        return Http(url)

    def _init_hash(self, algorithm: HashAlgorithm) -> Hash:
        return Hash(algorithm)

    def _init_link_generator(self) -> LinkGenerator:
        return LinkGenerator(hash=self._hash, password_1=self._password1)

    def _init_signature_checker(self) -> SignatureChecker:
        return SignatureChecker(
            hash_=self._hash, password1=self._password1, password2=self._password2
        )

    def _init_operation_state_checker(self) -> OperationStateChecker:
        return OperationStateChecker(
            merchant_login=self._merchant_login,
            hash=self._hash,
            password_2=self._password2,
        )

    @property
    def merchant_login(self) -> str:
        return self._merchant_login

    @property
    def algorithm(self) -> HashAlgorithm:
        return self._algorithm

    @property
    def is_test(self) -> bool:
        return self._is_test

    def __repr__(self):
        return f"{self.__class__.__name__}(is_test={self.is_test})"


class Robokassa(BaseRobokassa):
    def __init__(
        self,
        merchant_login: str,
        password1: str,
        password2: str,
        algorithm: HashAlgorithm = HashAlgorithm.md5,
        is_test: bool = False,
    ) -> None:
        super().__init__(
            merchant_login=merchant_login,
            password1=password1,
            password2=password2,
            algorithm=algorithm,
            is_test=is_test,
        )

    def generate_open_payment_link(
        self,
        out_sum: Union[float, int],
        default_prefix: str = "shp",
        result_url: Optional[str] = None,
        success_url: Optional[str] = None,
        success_url_method: Optional[HTTPMethod] = None,
        fail_url: Optional[str] = None,
        fail_url_method: Optional[HTTPMethod] = None,
        inv_id: int = 0,
        receipt: Optional[dict] = None,
        description: Optional[str] = None,
        recurring: bool = False,
        culture: Culture = Culture.RU,
        email: Optional[str] = None,
        expiration_date: Optional[datetime] = None,
        user_ip: Optional[str] = None,
        payment_methods: Optional[List[PaymentMethod]] = None,
        **kwargs,
    ) -> RobokassaResponse:
        """
        Create a link to payment page by common params with signature.
        Link of this method will look like:


        `https://auth.robokassa.ru/Merchant/Index.aspx?MerchantLogin=demo&OutSum=1&SignatureValue=2c113e992e2c985e43e`

        All link params user can see, but cannot edit them.
        If you want to hide these params you need to use by invoice ID method.


        :param out_sum:
        :param default_prefix: `shp` or `Shp` or `SHP` prefix for additional params
        :param result_url: ResultUrl2
        :param success_url: SuccessUrl2
        :param success_url_method: SuccessUrlMethod2
        :param fail_url: FailUrl2
        :param fail_url_method: FailUrlMethod2
        :param inv_id:
        :param description: Shop description
        :param kwargs: Any additional params without `shp_` prefix
        :return: Link to payment page
        """

        available_http_methods = (HTTPMethod.GET, HTTPMethod.POST, None)

        if (success_url is not None) != (success_url_method is not None):
            raise UnusedStrictUrlParameterError(
                "If you use URL, you also need to choose a HTTP method"
            )
        if (
            success_url_method not in available_http_methods
            or fail_url_method not in available_http_methods
        ):
            raise IncorrectUrlMethodError("You can use only GET or POST methods")

        params = RobokassaParams(
            is_test=self._is_test,
            merchant_login=self._merchant_login,
            out_sum=out_sum,
            inv_id=inv_id,
            receipt=receipt,
            description=description,
            recurring=recurring,
            payment_methods=payment_methods,
            success_url=success_url,
            success_url_method=success_url_method,
            fail_url=fail_url,
            fail_url_method=fail_url_method,
            result_url=result_url,
            default_prefix=default_prefix,
            culture=culture,
            email=email,
            user_ip=user_ip,
            expiration_date=expiration_date,
            additional_params=kwargs,
        )
        return self._link_generator.generate_open_payment_link(params=params)

    def generate_subscription_link(
        self,
        inv_id: int,
        previous_inv_id: int,
        out_sum: int,
        receipt: Optional[dict] = None,
    ) -> RobokassaResponse:
        params = RobokassaParams(
            is_test=self._is_test,
            merchant_login=self._merchant_login,
            out_sum=out_sum,
            inv_id=inv_id,
            previous_inv_id=previous_inv_id,
            receipt=receipt,
        )
        return self._link_generator.generate_subscription_payment_link(params=params)

    async def generate_protected_payment_link(
        self,
        invoice_type: InvoiceType,
        out_sum: float,
        merchant_comments: str,
        default_prefix: str = "shp",
        result_url: Optional[str] = None,
        success_url: Optional[str] = None,
        success_url_method: Optional[str] = None,
        fail_url: Optional[str] = None,
        fail_url_method: Optional[str] = None,
        inv_id: int = 0,
        receipt: Optional[dict] = None,
        description: Optional[str] = None,
        recurring: bool = False,
        culture: Culture = Culture.RU,
        email: Optional[str] = None,
        expiration_date: Optional[datetime] = None,
        **kwargs,
    ) -> RobokassaResponse:
        """
        Generate payment link using JWT
        https://docs.robokassa.ru/pay-interface/#jwt


        :param invoice_type: Type of invoice
        :type invoice_type: InvoiceType
        :param out_sum: Out sum
        :type out_sum: float
        :param merchant_comments: Description for merchant
        :type merchant_comments: str
        :param default_prefix: Prefix: shp, Shp or SHP. Default: shp (use without `_`)
        :type default_prefix: str
        :param result_url: url
        :type result_url:
        :param success_url: url
        :type success_url: str
        :param success_url_method: Method: GET, POST
        :type success_url_method:
        :param fail_url: url
        :type fail_url: str
        :param fail_url_method: Method: GET, POST
        :type fail_url_method:
        :param inv_id: Invoice ID
        :type inv_id: int
        :param receipt: Receipt. See: https://docs.robokassa.ru/fiscalization/
        :type receipt: dict
        :param description: Description of payment page
        :type description: str
        :param recurring: Is recurring payment? See: https://docs.robokassa.ru/recurring/
        :type recurring: bool
        :param culture: Language of payment page
        :type culture: Culture
        :param email: User email
        :type email: str
        :param expiration_date: Datetime obj when link expire
        :type expiration_date:
        :param kwargs: additional payments params without shp_.
        :type kwargs:
        :return: RobokassaResponse
        :rtype: RobokassaResponse
        """

        if self.is_test:
            raise RobokassaInterfaceError(
                "generate_protected_payment_link is unavailable in Test mode"
            )

        available_http_methods = ("GET", "POST", None)

        if (success_url is not None) != (success_url_method is not None):
            raise UnusedStrictUrlParameterError(
                "If you use URL, you also need to choose a HTTP method"
            )
        if (
            success_url_method not in available_http_methods
            or fail_url_method not in available_http_methods
        ):
            raise IncorrectUrlMethodError("You can use only GET or POST methods")

        params = RobokassaParams(
            is_test=self._is_test,
            merchant_login=self._merchant_login,
            out_sum=out_sum,
            inv_id=inv_id,
            receipt=receipt,
            description=description,
            recurring=recurring,
            merchant_comments=merchant_comments,
            invoice_type=invoice_type,
            success_url=success_url,
            success_url_method=success_url_method,
            fail_url=fail_url,
            fail_url_method=fail_url_method,
            result_url=result_url,
            default_prefix=default_prefix,
            culture=culture,
            email=email,
            expiration_date=expiration_date,
            additional_params=kwargs,
        )

        return await self._link_generator.generate_protected_payment_link(
            http=self._create_http(SERVICES_BASE_URL),
            params=params,
        )

    async def deactivate_invoice(
        self,
        inv_id: Optional[int] = None,
        encoded_id: Optional[str] = None,
        id: Optional[str] = None,
    ) -> None:
        """
        Deactivate invoice using invoiceId, Id or EncodedId

        :param inv_id: The account number specified by the seller when creating the link.
        :type inv_id:
        :param encoded_id: The last part of the invoice link.
        For example: https://auth.robokassa.ru/merchant/Invoice/6hucaX7-BkKNi4lyi-Iu2g.
        :type encoded_id: str
        :param id: The invoice ID is returned in response to an invoice request.
        :type id: str
        """

        if not inv_id and not encoded_id and not id:
            raise UnusedParameterError(
                "Provide inv_id or encoded_id or id of invoice for deactivation"
            )

        return await self._link_generator.deactivate_protected_payment_link(
            http=self._create_http(SERVICES_BASE_URL),
            merchant_login=self._merchant_login,
            inv_id=inv_id,
            encoded_id=encoded_id,
            id=id,
        )

    def is_redirect_valid(
        self,
        signature: str,
        out_sum: Union[str, int, float],
        inv_id: Optional[Union[str, int]],
        **kwargs,
    ) -> bool:
        """
        Check success or fail signature is valid.

        :param signature: Output signature
        :param out_sum:
        :param inv_id:
        :param kwargs: Additional params with `shp_` prefix
        :return: True if signature is valid, else False
        """

        return self._signature_checker.success_or_fail_url_signature_is_valid(
            success_signature=signature, out_sum=out_sum, inv_id=inv_id, **kwargs
        )

    def is_result_notification_valid(
        self,
        signature: str,
        out_sum: Union[str, int, float],
        inv_id: Optional[Union[str, int]],
        **kwargs,
    ) -> bool:
        """
        Check result signature is valid.

        :param signature: Output signature
        :param out_sum:
        :param inv_id:
        :param kwargs: Additional params with `shp_` prefix
        :return: True if signature is valid, else False
        """

        return self._signature_checker.result_url_signature_is_valid(
            result_signature=signature, out_sum=out_sum, inv_id=inv_id, **kwargs
        )

    async def get_payment_details(self, inv_id: int) -> PaymentDetails:
        """
        Get details of operation by invoice ID

        :param inv_id: Invoice ID
        :type inv_id: int
        :return: Details of operation
        :rtype: PaymentDetails
        """

        return await self._operation_state_checker.get_state(
            http=self._create_http(AUTH_BASE_URL), inv_id=inv_id
        )



from collections.abc import Sequence
from typing import Any, Dict, Optional, Union
from urllib.parse import quote, urlencode

from robokassa.connection import Http
from robokassa.exceptions import RobokassaInterfaceError
from robokassa.hash import Hash
from robokassa.jwt import JWT
from robokassa.types import RobokassaParams, RobokassaResponse, Signature


class LinkGenerator:
    def __init__(self, hash: Hash, password_1: str) -> None:
        self._static_url = "https://auth.robokassa.ru/Merchant/Index.aspx"
        self._password = password_1
        self._hash = hash

    def _create_signature(
        self,
        merchant_login: Optional[str] = None,
        out_sum: Optional[Union[float, str, int]] = None,
        inv_id: Optional[Union[str | int]] = None,
        receipt: Optional[dict] = None,
        result_url2: Optional[str] = None,
        success_url2: Optional[str] = None,
        success_url2_method: Optional[str] = None,
        fail_url2: Optional[str] = None,
        fail_url2_method: Optional[str] = None,
        additional_params: Optional[Dict[str, int | str | float]] = None,
    ) -> Signature:
        return Signature(
            hash_=self._hash,
            merchant_login=merchant_login,
            out_sum=out_sum,
            inv_id=inv_id,
            receipt=receipt,
            result_url2=result_url2,
            success_url2=success_url2,
            success_url2_method=success_url2_method,
            fail_url2=fail_url2,
            fail_url2_method=fail_url2_method,
            password=self._password,
            additional_params=additional_params,
        )

    def _sort_url_params(
        self, urls: Dict[str, str], success_url_method: str, fail_url_method: str
    ) -> Dict[str, str]:
        urls_plus_methods = {}
        for k, v in urls.items():
            if v is not None:
                urls_plus_methods[k] = f"{v}"
                if k == "SuccessUrl2":
                    method = ["SuccessUrl2Method", success_url_method]
                elif k == "FailUrl2":
                    method = ["FailUrl2Method", fail_url_method]
                else:
                    method = None
                if method is not None:
                    urls_plus_methods[method[0]] = method[1]
        return urls_plus_methods

    def _get_serialized_link_to_payment_page(self, url_params: Dict[str, Any]) -> str:
        return f"{self._static_url}?{urlencode(url_params)}"

    def _serialize_url_params(
        self, params: RobokassaParams, additional: Sequence
    ) -> dict:
        return {
            k: v
            for k, v in [
                ("MerchantLogin", params.merchant_login),
                ("OutSum", params.out_sum),
                ("InvId", params.inv_id),
                ("UserIp", params.user_ip),
                ("Receipt", params.receipt),
                ("Culture", params.culture),
                ("Description", params.description),
                ("ExpirationDate", params.expiration_date),
                ("Email", params.email),
                ("Recurring", "true" if params.recurring else False),
                *additional.items(),
                ("SignatureValue", params.signature_value),
                ("IsTest", int(params.is_test)),
            ]
            if v and v != "null" or (isinstance(v, int) and v == 0)
        }

    def _assemble_url(self, params: RobokassaParams) -> str:
        additional_params_for_url = sorted(
            [f"{k}={v}" for k, v in params.additional_params.items()]
        )

        urls_plus_methods = self._sort_url_params(
            {
                "ResultUrl2": params.result_url,
                "SuccessUrl2": params.success_url,
                "FailUrl2": params.fail_url,
            },
            params.success_url_method,
            params.fail_url_method,
        )
        url_params = self._serialize_url_params(params, urls_plus_methods)

        url = (
            f"{self._get_serialized_link_to_payment_page(url_params)}"
            f"&{'&'.join(additional_params_for_url)}"
        )

        return url

    def _escape_link(self, link: str) -> str:
        return quote(link, safe="")

    def _create_header_jwt(self) -> dict:
        algorithm = self._hash.algorithm.value.upper()

        return {"typ": "JWT", "alg": algorithm}

    def generate_open_payment_link(self, params: RobokassaParams) -> RobokassaResponse:
        signature = self._create_signature(
            merchant_login=params.merchant_login,
            out_sum=params.out_sum,
            inv_id=params.inv_id,
            receipt=params.receipt,
            result_url2=self._escape_link(params.result_url)
            if params.result_url
            else None,
            success_url2=self._escape_link(params.success_url)
            if params.success_url
            else None,
            success_url2_method=params.success_url_method,
            fail_url2=self._escape_link(params.fail_url) if params.fail_url else None,
            fail_url2_method=params.fail_url_method,
            additional_params=params.additional_params,
        )
        params.signature_value = signature.value

        return RobokassaResponse(url=self._assemble_url(params), params=params)

    def generate_subscription_payment_link(
        self, params: RobokassaParams
    ) -> RobokassaResponse:
        signature = self._create_signature(
            out_sum=params.out_sum,
            inv_id=params.inv_id,
            receipt=params.receipt,
            additional_params=params.additional_params,
        )
        params.signature_value = signature.value

        return RobokassaResponse(url=self._assemble_url(params), params=params)

    async def generate_protected_payment_link(
        self, http: Http, params: RobokassaParams
    ) -> RobokassaResponse:
        header = self._create_header_jwt()
        payload = self._serialize_url_params(
            params,
            {
                "InvoiceType": params.invoice_type,
                "MerchantComments": params.merchant_comments,
            },
        )

        del payload["IsTest"]

        signature = f"{params.merchant_login}:{self._password}"
        jwt = JWT(
            header=header, payload=payload, signature_key=signature, hash=self._hash
        ).create()

        async with http as conn:
            response = await conn.post("CreateInvoice", json=jwt)
            result = response.json()

            if result.get("isSuccess"):
                params.id = result["id"]
                return RobokassaResponse(url=result["url"], params=params)
            raise RobokassaInterfaceError("Failed to create link")

    async def deactivate_protected_payment_link(
        self,
        http: Http,
        merchant_login: str,
        encoded_id: Optional[str],
        id: Optional[str],
        inv_id: Optional[int],
    ) -> None:
        header = self._create_header_jwt()
        params = {
            "MerchantLogin": merchant_login,
            "InvId": inv_id,
            "Id": id,
            "EncodedId": encoded_id,
        }
        payload = {k: v for k, v in params.items() if v}
        signature = f"{merchant_login}:{self._password}"

        jwt = JWT(
            header=header, payload=payload, signature_key=signature, hash=self._hash
        ).create()

        async with http as conn:
            response = await conn.post("DeactivateInvoice", json=jwt)
            result = response.json()

            if not result.get("isSuccess"):
                raise RobokassaInterfaceError("Failed to deactivate invoice")



import _hashlib
import hashlib
from enum import Enum

from robokassa.exceptions import UnresolvedAlgorithmTypeError


class HashAlgorithm(Enum):
    md5 = "md5"
    ripemd160 = "ripemd160"
    sha1 = "sha1"
    sha256 = "sha256"
    sha384 = "sha384"
    sha512 = "sha512"


class Hash:
    def __init__(self, algorithm: HashAlgorithm) -> None:
        self.algorithm = algorithm

        if not isinstance(self.algorithm, HashAlgorithm):
            raise UnresolvedAlgorithmTypeError("Use HashAlgorithm class for that")

    def _encrypt_ripemd160(self, *args, **kwargs) -> _hashlib.HASH:
        h = hashlib.new("ripemd160")
        h.update(*args, **kwargs)
        return h

    def encrypt(self) -> _hashlib.HASH:
        if self.algorithm == HashAlgorithm.md5:
            result = hashlib.md5
        elif self.algorithm == HashAlgorithm.ripemd160:
            result = self._encrypt_ripemd160
        elif self.algorithm == HashAlgorithm.sha1:
            result = hashlib.sha1
        elif self.algorithm == HashAlgorithm.sha256:
            result = hashlib.sha256
        elif self.algorithm == HashAlgorithm.sha384:
            result = hashlib.sha384
        elif self.algorithm == HashAlgorithm.sha512:
            result = hashlib.sha512
        else:
            raise UnresolvedAlgorithmTypeError("Cannot define algorithm for hashing")

        return result

    def hash_data(self, data: str) -> str:
        data = data.encode()
        # str to bytes for hash

        hash = self.encrypt()

        return hash(data).hexdigest()
